"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.compileWat = exports.phs = exports.compile = void 0;
const scan_1 = require("../lib/scan");
const parse_1 = require("../lib/parse");
const context_1 = require("../lib/context");
// Import wabt promise
const wabtMod = require("wabt");
const wabtProm = wabtMod();
/**
 * Compile source code and instantiate module
 * @param src source code
 * @param importObject bindings imported from js
 * @param options unused for now
 */
async function compile(src, importObject = {}, options = {}) {
    // TODO bindings
    const ctx = (0, parse_1.default)((0, scan_1.default)(src, "inline"));
    if (!(ctx instanceof context_1.default))
        throw ctx;
    const wasm = await ctx.outWasm();
    const valid = WebAssembly.validate(wasm.buffer);
    if (!valid)
        throw new Error("WebAssembly.validate() failed");
    return await WebAssembly.instantiate(wasm.buffer, importObject);
}
exports.compile = compile;
;
// Tagged template string literal
/**
 *
 * @param {string[]} src - phs source code
 * @param {Array} bindings - js entities to be included in the module
 */
async function phs(src, ...bindings) {
    return await compile(src.reduce((acc, part, i) => `${acc} "phs_js_bind ${i}" ${part}`), bindings.reduce((acc, v, i) => ({ ...acc, [`phs_js_bind ${i}`]: v }), {}));
}
exports.phs = phs;
;
/*
Bindings: in this example it prompts user at compile time for this
const m = await phs`
    {${Number(prompt('enter a number'))} 10 + } $f =
    {} { f } $plus10 export
`;
*/
/**
 * @param {string} src - webasseembly text form source
 * @param {*} importObject
 * @returns {WebAssembly.Instance}
 */
async function compileWat(src, importObject = {}) {
    const wabt = await wabtProm;
    const mod = wabt.parseWat("inline", src, {
        exceptions: true,
        mutable_globals: true,
        sat_float_to_int: true,
        sign_extension: true,
        simd: true,
        threads: true,
        multi_value: true,
        tail_call: true,
        bulk_memory: true,
        reference_types: true,
        annotations: true,
        gc: true,
    });
    // Validate
    mod.validate();
    const bin = mod.toBinary({ log: true });
    const valid = WebAssembly.validate(bin.buffer);
    if (!valid)
        return console.error("wasm invalid!", valid);
    return await WebAssembly.instantiate(bin.buffer, importObject);
}
exports.compileWat = compileWat;
;
//# sourceMappingURL=inline.js.map