#!/usr/bin/env node
import yargs = require('yargs');
import cp = require('child_process');

import runShell from './tools/shell';
import compileFile from './tools/file';

yargs
    .scriptName('phc')
    .usage('$0 <command> [args]')
    .option('verbose', {
        describe: 'include verbose output',
        type: 'boolean',
        default: false,
        alias: 'v',
    })
    .command(['shell [options]', '*'], 'run interactive shell',
        yargs =>
            yargs.options({
                'lex': {
                    describe: 'debug lexer tokens',
                    type: 'boolean',
                    default: false,
                    alias: 'l' },
            }),
        argv =>
            runShell(argv.lex, argv.verbose))
    .command('file <name> [options]', 'compile a file to WAT',
        yargs => yargs
            .positional('name', {
                describe: 'name of the file to open',
                type: 'string',
            })
            .options({
                'track-time': {
                    describe: 'track time spent compiling',
                    type: 'boolean',
                    default: true,
                    alias: 't',
                },
                'fast' : {
                    describe: 'skip validation and pretty-print steps',
                    type: 'boolean',
                    default: false,
                },
                'folding' : {
                    describe: 'use folding/s-expr WAT syntax',
                    type: 'boolean',
                    default: false,
                },
                'optimize' : {
                    describe: 'pass compiled output through binaryen optimizer',
                    default: false,
                    alias: 'O',
                },
            }),
        argv =>
            compileFile(argv.name, argv['track-time'], argv.fast, argv.folding, argv.optimize))
    // .help()
    .argv;